from .lstm import LSTM
from .cnn import cnn_1d
from .attention_encoder import enc_classifier, encoder
from torch import nn

_models = {"lstm": LSTM, "cnn1d": cnn_1d, "encoder": enc_classifier}


def load_model(key):
    if key in _models:
        return _models[key]
    else:
        raise ValueError(f'Invalid model architecture name, "{key}"')


class head(nn.Module):
    def __init__(self, inp_size, output_size=None, hidden=None, **kwargs):
        super(head, self).__init__()
        if hidden:
            nodes = [inp_size] + hidden + [output_size]
        else:
            nodes = [inp_size] + [output_size]
        self.linear = []
        # print(nodes)
        for i in range(len(nodes) - 1):
            self.linear.append(nn.ReLU())
            self.linear.append(nn.Linear(int(nodes[i]), int(nodes[i + 1])))

        self.linear = nn.ModuleList(self.linear)

    def forward(self, x):
        for layer in self.linear:
            x = layer(x)
        return x


class model_tot(nn.Module):
    def __init__(
        self,
        feat,
        linear_head=None,
        projector=None,
        # feat2=None,
        # linear_head2=None,
        # projector2=None,
    ):
        super(model_tot, self).__init__()
        self.feat = feat
        self.linear_head = linear_head
        self.projector = projector
        # self.feat2 = feat2
        # self.linear_head2 = linear_head2
        # self.projector2 = projector2

    def forward(self, batch):
        x = self.feat(**batch)
        output = {}
        if self.linear_head:
            lh = self.linear_head(x)
            output["lin_head"] = lh
        if self.projector:
            proj = self.projector(x)
            output["proj"] = proj

            # return_data["proj"] = proj
        # if self.feat2:
        #     x2 = self.feat2(**batch2)
        #     lh2 = self.linear_head2(x2)
        #     return_data["lh2"] = lh2
        #     if self.projector2:
        #         proj2 = self.projector2(x2)
        #         return_data["proj2"] = proj2
        return output


class contrastive_model(nn.Module):
    def __init__(self, model1, model2) -> None:
        super().__init__()
        self.model1 = model1
        self.model2 = model2

    def forward(self, batch):
        out1 = self.model1(batch[0])
        # print(out1)
        out2 = self.model2(batch[1])
        # print(out2)
        return out1, out2



import torch
from torch import nn

def conv3x3(in_planes, out_planes, stride=1):
    return nn.Conv2d(in_planes, out_planes, kernel_size=3, stride=stride,
                     padding=1, bias=False)

def conv1x1(in_planes, out_planes, stride=1):
    return nn.Conv2d(in_planes, out_planes, kernel_size=1, stride=stride,
                     bias=False)

class basicBlock(nn.Module):
    expansion = 1

    def __init__(self, in_planes, planes, stride=1, down_sample=None):
        super(basicBlock, self).__init__()
        self.conv1 = conv3x3(in_planes, planes, stride)
        self.bn1 = nn.BatchNorm2d(planes)
        self.relu = nn.ReLU(inplace=True)
        self.conv2 = conv3x3(planes, planes)
        self.bn2 = nn.BatchNorm2d(planes)
        self.down_samples = down_sample
        self.stride = stride

    def forward(self, x):
        identity = x
        if self.down_samples is not None:
            identity = self.down_sample(identity)

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)
        out += identity
        out = self.relu(out)
        
        return out
        
class bottleNeck(nn.Module):
    expansion = 4

    def __init__(self, in_planes, planes, stride=1, down_sample=None):
        super(bottleNeck, self).__init__()
        self.conv1 = conv1x1(in_planes, planes)
        self.bn1 = nn.BatchNorm2d(planes)
        self.conv2 = conv3x3(planes, planes, stride)
        self.bn2 = nn.BatchNorm2d(planes)
        self.conv3 = conv1x1(planes, planes * self.expansion)
        self.bn3 = nn.BatchNorm2d(planes)

        self.relu = nn.ReLU(inplace=True)

        self.down_sample = down_sample
        self.stride = stride

    def forward(self, x):
        identity = x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)
        out = self.relu(out)

        out = self.conv3(out)
        out = self.bn3(out)

        if self.down_sample is not None:
            identity = self.down_sample(x)(
        
        out += identity
        out = self.relu(out)

        return out
    
class ResNet(nn.Module):
    def __init__(self, block, layers, num_classes, zero_init_residual=False):
        super(ResNet, self).__init__()
        self.inplanes = 64₩
        self.conv1 = nn.Conv2d(3, 64, kernel_size=7, stride=2, padding=3, bias=False)
        self.bn1 = nn.BatchNorm2d(64)
        self.relu = nn.ReLU(inplace=True)
        self.maxpool = nn.MaxPool2d(kernel_size=3, stride=2, padding=1)
            
    def _make_layer(self, block, planes, blocks, stride=1):
        down_sample = None
        if stride != 1 or self.inplanes != planes * block.expansion:
            down_sample = nn.Sequential(
                conv1x1(self.inplanes, planes * block.expansion, stride),
                nn.BatchNorm2d(planes * block.expansion)
            )


# %%
from torch import nn
import torch

# def cnn1d(in_ch, out_ch, stride=1, padding='same'):


class _cnn1d_block(nn.Module):
    def __init__(
        self,
        in_ch: int,
        out_ch: int,
        dropout_rate: float,
        down_scale: bool = False,
        kernel_size: int = 3,
    ) -> None:
        super(_cnn1d_block, self).__init__()

        stride = 2 if down_scale else 1

        self.cnn1 = nn.Conv1d(
            in_ch, out_ch, kernel_size, stride=stride, padding=1, bias=False
        )
        self.bn1 = nn.BatchNorm1d(out_ch)
        self.cnn2 = nn.Conv1d(
            out_ch, out_ch, kernel_size, stride=1, padding=1, bias=False
        )
        self.bn2 = nn.BatchNorm1d(out_ch)
        self.dropout = nn.Dropout(dropout_rate)

        self.relu = nn.ReLU(inplace=True)
        self.downscale = None
        if down_scale:
            self.downscale = nn.Conv1d(in_ch, out_ch, kernel_size=1, stride=stride)

    def forward(self, x):
        identity = x
        if self.downscale is not None:
            identity = self.downscale(identity)
        out = self.cnn1(x)
        out = self.bn1(out)
        out = self.relu(out)
        out = self.dropout(out)

        out = self.cnn2(out)
        out = self.bn2(out)
        out += identity
        out = self.relu(out)

        return out


class cnn_1d(nn.Module):
    def __init__(
        self,
        vocab_size: int,
        emb_dim: int,
        num_layer: int,
        channels: list,
        ini_channel: int = 32,
        dropout_rate: float = 0.3,
        **kwargs
    ) -> None:
        super(cnn_1d, self).__init__()

        # assert len(num_layers) == len(channels), "Dim of layers and channels must be equal"
        self.embed = nn.Embedding(vocab_size, emb_dim, padding_idx=0)

        self.curr_channel = ini_channel
        self.conv1 = nn.Conv1d(
            emb_dim,
            self.curr_channel,
            kernel_size=5,
            stride=1,
            padding="same",
            bias=False,
        )
        self.bn1 = nn.BatchNorm1d(self.curr_channel)
        self.relu = nn.ReLU(inplace=True)
        self.maxpool = nn.MaxPool1d(kernel_size=3, stride=1, padding=1)
        self.avgpool = nn.AdaptiveAvgPool1d(1)
        # self.fc = nn.Linear(channels[-1], n_cls)

        self.stacked_layers = []
        for channel in channels:
            self.stacked_layers += self._block_stack(num_layer, channel, dropout_rate)
        self.stacked_layers = nn.Sequential(*self.stacked_layers)
        self.out_dim = channels[-1]

    def _block_stack(self, num_layer: int, channel: int, dropout_rate: float):
        layers = []
        layers.append(_cnn1d_block(self.curr_channel, channel, dropout_rate, True))
        for _ in range(1, num_layer):
            layers.append(_cnn1d_block(channel, channel, dropout_rate))
        self.curr_channel = channel
        return layers

    def forward(self, x, **kwargs):
        x = self.embed(x)
        x = x.transpose(1, 2)
        x = self.conv1(x)
        x = self.bn1(x)
        x = self.relu(x)
        x = self.maxpool(x)
        x = self.stacked_layers(x)
        x = self.avgpool(x)
        x = x.view(x.size(0), -1)

        return x



from torch import nn
import torch
from torch.nn.utils.rnn import pack_padded_sequence, pad_packed_sequence


class LSTM(nn.Module):
    def __init__(
        self,
        vocab_size,
        emb_dim,
        n_hidden,
        n_RNN_layers,
        device,
        dropout_rate=0.5,
        bidirectional=False,
        **kwargs
    ):
        super(LSTM, self).__init__()
        self.embed = nn.Embedding(vocab_size, emb_dim, padding_idx=0)
        self.rnn = nn.LSTM(
            input_size=emb_dim,
            hidden_size=n_hidden,
            num_layers=n_RNN_layers,
            batch_first=True,
            dropout=dropout_rate,
            bidirectional=bidirectional,
        )
        self.out_dim = n_hidden * 2 if bidirectional else n_hidden
        # self.fc = nn.Linear(self.out_size, n_cls)
        self.device = device

    def forward(self, x, seq_len, **kwargs):
        emb = self.embed(x)

        packed_x = pack_padded_sequence(
            emb, seq_len.cpu(), batch_first=True, enforce_sorted=False
        )
        rnn_out, _ = self.rnn(packed_x)
        rnn_out, s_len = pad_packed_sequence(rnn_out, batch_first=True)
        s_idx = (s_len.to(self.device) - 1).unsqueeze(1).unsqueeze(2)
        s_idx = s_idx.expand(s_idx.size(0), 1, rnn_out.size(2))
        gather_res = torch.gather(rnn_out, 1, s_idx)
        gather_res = torch.squeeze(gather_res)
        # return self.fc(gather_res)
        return gather_res



from setuptools import setup, find_packages

setup(
    name="conACP-pred",
    version="0.0.1",
    description="Screening anticancer peptide from amino acid sequence data",
    author="zolee",
    author_email="blee.inf@gmail.com",
    url="https://github.com/bzlee-bio/con_ACP",
    install_requires=["torch"],
    packages=find_packages(exclude=[]),
    keywords=["anticancer peptide", "sequence data", "deep learning"],
    python_requires=">=3.6",
    package_data={},
    zip_safe=False,
    classifiers=[
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
    ],
)



import torch
from torch import nn, Tensor


class multihead_attention(nn.Module):
    def __init__(self, h, d_model, dropout=0.1):
        super(multihead_attention, self).__init__()
        assert d_model % h == 0
        self.w_q = nn.Linear(d_model, d_model)
        self.w_k = nn.Linear(d_model, d_model)
        self.w_v = nn.Linear(d_model, d_model)
        self.w_o = nn.Linear(d_model, d_model)

        self.h = h
        self.d_k = d_model // h

        self.attn_prob = None

        if dropout is not None:
            self.dropout = nn.Dropout(p=dropout)

    def attention(self, q: Tensor, k: Tensor, v: Tensor, mask=None, dropout=None):
        # q, k (..., seq_len, d_k)
        # v (..., seq_len, d_v)
        attn_score = (
            torch.matmul(q, k.transpose(-1, -2)) / k.size(-1) ** 0.5
        )  # QK^T/sqrt(d_k)
        if mask is not None:
            attn_score = attn_score.masked_fill(mask == 0, -1e9)
        attn_prob = torch.softmax(attn_score, dim=-1)
        if dropout is not None:
            attn_prob = dropout(attn_prob)
        return torch.matmul(attn_prob, v), attn_prob

    def forward(self, q, k, v, mask=None):
        # q, k, v (..., seq_len, d_model)
        n_batch = q.size(0)
        if mask is not None:
            mask = mask.unsqueeze(1)

        q = self.w_q(q)
        k = self.w_k(k)
        v = self.w_v(v)
        q = q.view(n_batch, -1, self.h, self.d_k).transpose(1, 2)
        k = k.view(n_batch, -1, self.h, self.d_k).transpose(1, 2)
        v = v.view(n_batch, -1, self.h, self.d_k).transpose(1, 2)

        x, self.attn_prob = self.attention(q, k, v, mask)
        x = x.transpose(1, 2).contiguous().view(n_batch, -1, self.h * self.d_k)
        return self.w_o(x)


class position_wise_feedforward(nn.Module):
    def __init__(self, d_model, d_ff, dropout=0.1) -> None:
        super(position_wise_feedforward, self).__init__()
        self.lin1 = nn.Linear(d_model, d_ff)
        self.lin2 = nn.Linear(d_ff, d_model)
        self.dropout = None
        if dropout is not None:
            self.dropout = nn.Dropout(dropout)
        self.relu = nn.ReLU(inplace=True)

    def forward(self, x):
        x = self.lin1(x)
        x = self.relu(x)
        if self.dropout is not None:
            x = self.dropout(x)
        return self.lin2(x)


class residual_connect(nn.Module):
    def __init__(self, d_model, dropout=0.1) -> None:
        """Residual connection & layer normalization

        Args:
            inp_shape (list or torch.size): _description_
        """
        super(residual_connect, self).__init__()
        self.layer_norm = nn.LayerNorm(d_model)
        self.dropout = None
        if dropout is not None:
            self.dropout = nn.Dropout(dropout)

    def forward(self, x, sub_layer):
        identity = x
        x = sub_layer(self.layer_norm(x))
        x = self.layer_norm(x)
        if self.dropout is not None:
            x = self.dropout(x)

        return x + identity


class encoder_layer(nn.Module):
    def __init__(self, h, d_model, d_ff, dropout) -> None:
        super(encoder_layer, self).__init__()
        self.self_attn = multihead_attention(h, d_model, dropout)
        self.pwff = position_wise_feedforward(
            d_model=d_model, d_ff=d_ff, dropout=dropout
        )
        self.dropout = dropout

        self.res_conn1 = residual_connect(d_model=d_model, dropout=dropout)
        self.res_conn2 = residual_connect(d_model=d_model, dropout=dropout)

    def forward(self, x, mask):
        x = self.res_conn1(x, lambda x: self.self_attn(x, x, x, mask))
        x = self.res_conn2(x, self.pwff)
        return x


class encoder(nn.Module):
    def __init__(self, num_encoder, h, d_model, d_ff, dropout) -> None:
        super().__init__()
        self.encoders = nn.ModuleList(
            [
                encoder_layer(h=h, d_model=d_model, d_ff=d_ff, dropout=dropout)
                for _ in range(num_encoder)
            ]
        )
        self.layer_norm = nn.LayerNorm(d_model)

    def forward(self, x, mask=None):
        for layer in self.encoders:
            x = layer(x, mask)
        return self.layer_norm(x)


class enc_classifier(nn.Module):
    def __init__(self, n_encoder, h, emb_dim, d_ff, dropout_rate, vocab_size, **kwargs):
        d_model = emb_dim
        super(enc_classifier, self).__init__()
        self.enc = encoder(
            num_encoder=n_encoder,
            h=h,
            d_model=d_model,
            d_ff=d_ff,
            dropout=dropout_rate,
        )
        self.embed = nn.Embedding(vocab_size, d_model, padding_idx=0)
        self.pos_embed = nn.Embedding(252, d_model, padding_idx=0)
        self.dropout = nn.Dropout(dropout_rate)
        self.idx = torch.tensor(list(range(250)), dtype=torch.int32)
        self.out_dim = d_model

    def forward(self, x, seq_idx, mask=None, **kwargs):
        x_pos = self.pos_embed(seq_idx)
        x = self.embed(x)
        x += x_pos
        x = self.enc(x, mask)
        return x[:, 0, :].squeeze()  # classification task



# %%
import json
import torch

from torch import nn
from torch.utils.data import DataLoader
import util.data as Data
from util.args_inference import arg_parser
import copy


# %%
def predict(fasta_file, model_type="ACP_Mixed_80", device="cpu", batch_size=64):
    """
    Predict anticancer peptides for a given fasta file using a specified model.

    Parameters:
    fasta_file (str): Path to the fasta file for which the prediction is to be made.
    model_type (str, optional): Type of the model to be used for prediction. Options include "ACP-Mixed-80", "ACP2.0 main", "ACP2.0 alternative", "ACP500+ACP164", "ACP500+ACP2710", and "LEE+Independent". Default is "ACP_Mixed_80".
    device (str, optional): Device to be used for prediction. It can be either "cpu" or "gpu". Default is "cpu".
    batch_size (int, optional): Number of samples per prediction. Default is 64.

    Returns:
    list: A list of lists. Each inner list contains the id, probability of ACPs, and prediction for a sample (ACP or -). For prediction "ACP", the peptide is predicted to be an anticancer peptide. For prediction "-", the peptide is predicted to be a non-anticancer peptide.
    """
    assert model_type in [
        "ACP-Mixed-80",
        "ACP2.0 main",
        "ACP2.0 alternative",
        "ACP500+ACP164",
        "ACP500+ACP2710",
        "LEE+Independent",
    ], "The provided model_type is not supported."
    assert device in ["cpu", "gpu"], "The provided device is not supported."
    assert batch_size > 0, "The batch size should be greater than 0."

    args = AttrDict(
        {
            "model_type": model_type,
            "device": device,
            "batch_size": batch_size,
            "input": fasta_file,
        }
    )

    # args = arg_parser()
    model_id = f"./ckpt/{args.model_type}"
    with open(f"{model_id}_best.json", "rt") as f:
        args += json.load(f)

    if args.model_type == "ACP2_main":
        from model_old import load_model, head, model_tot, contrastive_model
    else:
        from model import load_model, head, model_tot, contrastive_model

    args.device = torch.device("cpu") if args.device == "cpu" else torch.device("gpu")
    args.original_AA_tok_len = copy.deepcopy(args.AA_tok_len)

    start_token = (
        True
        if ((args.model == "encoder") & (args.dataset != "ACP2_main"))
        | (args.dataset == "ACP2_alter")
        else False
    )

    if args.model_type == "ACP2_main":
        tr_file = "./dataset/ACPred-LAF/ACP2_main_train.csv"
        test_file = "./dataset/ACPred-LAF/ACP2_main_test.csv"
        max_len = (51, 50)
    elif args.model_type == "ACP2_alter":
        tr_file = "./dataset/ACPred-LAF/ACP2_alternate_train.csv"
        test_file = "./dataset/ACPred-LAF/ACP2_alternate_test.csv"
        max_len = (51, 50)
    elif args.model_type == "LEE_Indep":
        tr_file = "./dataset/ACPred-LAF/LEE_Dataset.csv"
        test_file = "./dataset/ACPred-LAF/Independent dataset.csv"
        max_len = (96, 95)
    elif args.model_type == "ACP500_ACP164":
        tr_file = "./dataset/ACPred-LAF/ACP_FL_train_500.csv"
        test_file = "./dataset/ACPred-LAF/ACP_FL_test_164.csv"
        max_len = (207, 206)
    elif args.model_type == "ACP500_ACP2710":
        tr_file = "./dataset/ACPred-LAF/ACPred-Fuse_ACP_Train500.csv"
        test_file = "./dataset/ACPred-LAF/ACPred-Fuse_ACP_Test2710.csv"
        max_len = (207, 206)
    elif args.model_type == "ACP_Mixed_80":
        tr_file = "./dataset/ACPred-LAF/ACP-Mixed-80-train.csv"
        test_file = "./dataset/ACPred-LAF/ACP-Mixed-80-test.csv"
        max_len = (208, 207)
    else:
        raise ValueError("Correct model type is not provided.")

    x_tr, y_tr, x_val, y_val = Data.raw_data_read(
        tr_file, args.val_fold, seed=args.seed
    )
    x_test, y_test = Data.raw_data_read(test_file)
    start_token = (
        True
        if ((args.model == "encoder") & (args.dataset != "ACP2_main"))
        | (args.dataset == "ACP2_alter")
        else False
    )

    data_t1 = Data.dataset((x_tr, y_tr), 1, start_token=start_token)
    data_t2 = Data.dataset((x_tr, y_tr), 2, start_token=start_token)
    data_train = Data.pretrain_dataset(data_t1, data_t2)

    data_v1 = Data.dataset((x_val, y_val), 1, start_token=start_token)
    data_v2 = Data.dataset((x_val, y_val), 2, start_token=start_token)
    data_valid = Data.pretrain_dataset(data_v1, data_v2)

    data_te1 = Data.dataset((x_test, y_test), 1, start_token=start_token)
    data_te2 = Data.dataset((x_test, y_test), 2, start_token=start_token)
    data_test = Data.pretrain_dataset(data_te1, data_te2)

    if args.model == "lstm":
        collate_fn = Data.collate_fn_lstm(args.contrastive)
        head_inp_size = args.n_hidden * 2 if args.bidirectional else args.n_hidden
    elif args.model == "encoder":
        collate_fn = Data.collate_fn_encoder(args.contrastive)
        head_inp_size = args.emb_dim
    elif args.model == "cnn1d":
        collate_fn = Data.collate_fn_cnn(max_len, args.contrastive)
        head_inp_size = args.channels[-1]

    ## inference fasta read
    inf_data = Data.raw_data_read_fasta(args.input)

    # sequence length check
    for seq, id in zip(inf_data[0], inf_data[2]):
        assert (
            len(seq) < max_len[0]
        ), f"Sequence {id} exceeds maximum input sequence length (Maximum sequence length is {max_len[0]-1})"

    data_inf1 = Data.dataset((inf_data[0], inf_data[1]), 1, start_token=start_token)
    data_inf2 = Data.dataset((inf_data[0], inf_data[1]), 2, start_token=start_token)
    data_inf = Data.pretrain_dataset(data_inf1, data_inf2)
    inference_loader = DataLoader(
        data_inf, batch_size=args.batch_size, shuffle=False, collate_fn=collate_fn
    )
    inference_loader = Data.device_DataLoader(
        inference_loader, args.device, args.contrastive
    )

    ## model load
    if args.original_AA_tok_len == 1:
        args.AA_tok_len = 1
        args.vocab_size = data_t1.vocab_size()
        l_feature1 = load_model(args.model)(**args)
        l_proj1 = head(
            inp_size=head_inp_size,
            output_size=head_inp_size / 4,
            hidden=[head_inp_size / 2],
        )
        l_lin_head1 = head(inp_size=head_inp_size, output_size=2)
        classifier = model_tot(
            feat=l_feature1, linear_head=l_lin_head1, projector=l_proj1
        )

        classifier.load_state_dict(
            torch.load(f"{model_id}_best.pt", map_location=args.device)
        )
        args.model_id = 0
    else:
        args.AA_tok_len = 2
        args.vocab_size = data_t2.vocab_size()
        l_feature2 = load_model(args.model)(**args)
        l_proj2 = head(
            inp_size=head_inp_size,
            output_size=head_inp_size / 4,
            hidden=[head_inp_size / 2],
        )
        l_lin_head2 = head(inp_size=head_inp_size, output_size=2)
        classifier = model_tot(
            feat=l_feature2, linear_head=l_lin_head2, projector=l_proj2
        )

        classifier.load_state_dict(
            torch.load(f"{model_id}_best.pt", map_location=args.device)
        )
        args.model_id = 1

    classifier.eval()
    with torch.no_grad():
        for i, batch in enumerate(inference_loader):
            logit = classifier(batch[args.model_id])["lin_head"].squeeze()
            pred = torch.argmax(logit, dim=1).squeeze()

            if i == 0:
                pred_total = torch.argmax(logit, dim=1).squeeze()
                logit_prob = nn.Softmax(1)(logit)[:, 1]
            else:
                pred_total = torch.concat(
                    [pred_total, torch.argmax(logit, dim=1).squeeze()]
                )
                logit_prob = torch.concat([logit_prob, nn.Softmax(1)(logit)[:, 1]])
    return_res = []
    for id, logit, pred in zip(inf_data[2], logit_prob, pred_total):
        _pred = "ACP" if pred == 1 else "-"
        return_res.append([id, logit.item(), _pred])
    return return_res
